#ifndef SOCKEV_H
#define SOCKEV_H
#include <libporty/net/network.h>


void sockev_init(void);
void sockev_uninit(void);

typedef int sockev_t;
#define SOCKEV_INVALID -1
#define SOCKEV_NEW -128

typedef void sockev_cb_none(sockev_t id);
typedef int  sockev_cb_ret(sockev_t id);
typedef void sockev_cb_data(sockev_t id, char *buff, int len);

typedef struct sockev_hooks_s {
	sockev_cb_none *open;     /* called once, when reading and/or writing is enabled (connection established); should allocate userdata */
	sockev_cb_none *close;    /* called once, when socket is fully closed; must free userdata */
	sockev_cb_none *error;    /* called on error condition (typically close will be called after error returns) */

	sockev_cb_ret  *readable; /* called whenever pending data is detected on the socket; if it's NULL or returns 0, a read is performed and read callback will be called */
	sockev_cb_ret  *writable; /* called whenever socket can receive write;  if it's NULL or returns 0, next chunk of the output buffer will be written */

	sockev_cb_data *read;     /* called if readable hook didn't catch the request */
	sockev_cb_none *readeof;  /* called when eof is read (read half of the connection is closed) */
} sockev_hooks_t;

/* returns number of events */
int sockev_poll(int timeout_ms);

/* write binary content into a socket's write buffer */
void sockev_write(sockev_t id, const char *buff, int len);

/* write string into a socket's write buffer */
void sockev_puts(sockev_t id, const char *str);

/* printf content into a socket's write buffer */
void sockev_printf(sockev_t id, const char *format, ...);

/* read a socket (useful from the readable hook) */
int sockev_read(sockev_t id, char *buff, int len);


/* shut down the socket in one or both directions using P_SHUT_* */
void sockev_shutdown(sockev_t id, int direction);

/* close an existing connection */
#define sockev_close(id)   sockev_shutdown(id, P_SHUT_RDWR)

/* add hooks on a socket */
sockev_t sockev_cfg_hooks(sockev_t id, const sockev_hooks_t *hooks);

typedef enum sockev_flags_e {
	SEVFLG_NO_HALF_CLOSE = 1       /* close conneciton in both directions if it is closed in any direction */
} sockev_flags_t;

/* set flags on a socket */
sockev_t sockev_cfg_flags(sockev_t id, sockev_flags_t flags);

/* set and retrieve userdata assocated with a socket */
void sockev_set_userdata(sockev_t id, void *ud);
void *sockev_get_userdata(sockev_t id);

/* disable/enable read or write for flow control - if read or write is
   non-zero, enable polling for read and/or write; returns id */
sockev_t sockev_set_watch(sockev_t id, int read, int write);

/* return remote address information */
int sockev_get_remote(sockev_t id, P_ipv4_addr_t *ip, const char **hostname, int *port);

/* return number of bytes in write buffer */
int sockev_get_wbuf(sockev_t id);

/* === filters === */
typedef enum sockev_f_event_e {
	SOCKEV_F_INIT,                /* Initialize; data holds init data; never call next */
	SOCKEV_F_IN,                  /* *data/len read from the network; call sockev_filter_next to pass on data; may modify data */
	SOCKEV_F_OUT,                 /* *data/len is being written to the network; call sockev_filter_next to pass on data; MUST NOT MODIFY data */
	SOCKEV_F_OPEN,                /* connection established (bidir) */
	SOCKEV_F_CLOSE                /* connection closed (both dirs) - UNINIT! */
} sockev_f_dir_t;

/* filter callback type */
typedef void (*sockev_filter_cb)(sockev_t id, void **sockdata, sockev_f_dir_t event, void *data, int len, int cookie);

typedef struct sockev_filters_s {
	sockev_filter_cb filter;
	void *filter_init_data;
} sockev_filters_t;

/* configure socket to have a filter chain; filters is a NULL terminated array,
   filters listed in application->network direction */
sockev_t sockev_cfg_filters(sockev_t id, const sockev_filters_t *filters);

/* configure everything at once */
#define sockev_cfg(id, hooks, flags, filters) sockev_cfg_filters(sockev_cfg_hooks(sockev_cfg_flags((id), (flags)), (hooks)), filters)

/* create a new socket and configure everything */
#define sockev_new(hooks, flags, filters) sockev_cfg(SOCKEV_NEW, (hooks), (flags), (filters))

#endif
